@@ -1,7 +1,6 @@
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 1996  Erich Boleyn  <erich@uruk.org>
- *  Copyright (C) 1999  Free Software Foundation, Inc.
+ *  Copyright (C) 1999, 2001  Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -257,6 +256,8 @@ ext2fs_mount (void)
 
   if ((((current_drive & 0x80) || (current_slice != 0))
        && (current_slice != PC_SLICE_TYPE_EXT2FS)
+       && (current_slice != PC_SLICE_TYPE_LINUX_RAID)
+       && (! IS_PC_SLICE_TYPE_BSD_WITH_FS (current_slice, FS_EXT2FS))
        && (! IS_PC_SLICE_TYPE_BSD_WITH_FS (current_slice, FS_OTHER)))
       || part_length < (SBLOCK + (sizeof (struct ext2_super_block) / DEV_BSIZE))
       || !devread (SBLOCK, 0, sizeof (struct ext2_super_block),
@@ -431,16 +432,12 @@ ext2fs_read (char *buf, int len)
       if (size > len)
 	size = len;
 
-#ifndef STAGE1_5
       disk_read_func = disk_read_hook;
-#endif /* STAGE1_5 */
 
       devread (map * (EXT2_BLOCK_SIZE (SUPERBLOCK) / DEV_BSIZE),
 	       offset, size, buf);
 
-#ifndef STAGE1_5
       disk_read_func = NULL;
-#endif /* STAGE1_5 */
 
       buf += size;
       len -= size;
