@@ -1,7 +1,7 @@
 /* shared.h - definitions used in all GRUB-specific code */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 1999, 2000, 2001 Free Software Foundation, Inc.
+ *  Copyright (C) 1999,2000,2001,2002  Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -548,6 +548,7 @@ typedef enum
   ERR_PRIVILEGED,
   ERR_NEED_SERIAL,
   ERR_NO_DISK_SPACE,
+  ERR_NUMBER_OVERFLOW,
 
   MAX_ERR_NUM
 } grub_error_t;
@@ -608,6 +609,11 @@ extern char *password;
 extern password_t password_type;
 extern int auth;
 extern char commands[];
+
+/* For `more'-like feature.  */
+extern int max_lines;
+extern int count_lines;
+extern int use_pager;
 #endif
 
 #ifndef NO_DECOMPRESSION
@@ -839,7 +845,8 @@ void stop_floppy (void);
 #define BUILTIN_MENU		0x2	/* Run in the menu.  */
 #define BUILTIN_TITLE		0x4	/* Only for the command title.  */
 #define BUILTIN_SCRIPT		0x8	/* Run in the script.  */
-#define BUILTIN_HIDDEN		0x10	/* Don't print command on booting. */
+#define BUILTIN_NO_ECHO		0x10	/* Don't print command on booting. */
+#define BUILTIN_HELP_LIST	0x20	/* Show help in listing.  */
 
 /* The table for a builtin.  */
 struct builtin
